from huggingface_hub import HfApi, CommitOperationAdd, CommitOperationDelete
from huggingface_hub import create_repo,metadata_update,cached_assets_path
from typing import List, Optional, Union
import json
import os
import pathlib
import logging 

class Repository():
    def __init__(self,name:str,user_or_organization:Optional[str]=None,token:Optional[str]=None,private:bool=False) -> None:
        
        self.name = f"{user_or_organization}/{name}" if user_or_organization else name
        self.api = HfApi(token=token)

        try:
            self.url = str(create_repo(self.name,token=token,private=private))
            self.name = self.url.replace(f"{self.api.endpoint}/","")
            # Uplaod the config file
            config_path = cached_assets_path("llm-rs",namespace=self.name) / "config.json"
            with open(config_path,"w") as f:
                f.write(json.dumps({"repo_type":"GGML"}))
            self.api.create_commit(
                self.name,
                operations=[CommitOperationAdd(path_in_repo="config.json", path_or_fileobj=config_path)],
                commit_message="Auto initialized repo via llm-rs",)
            
            metadata={}
            metadata["pipeline_tag"]= "text-generation"
            metadata["tags"] = ["llm-rs","ggml"]
            metadata_update(self.name, metadata, token=token)
            
        except Exception as e:
            self.url = create_repo(self.name,token=token,private=private,exist_ok=True)
            self.name = self.url.replace(f"{self.api.endpoint}/","")

    def upload(self,model_file:Union[str,os.PathLike],delete_old:bool=True):
        # search for the metadata file
        path = pathlib.Path(model_file)
        metadata_path = path.with_suffix(".meta")

        if not path.exists():
            raise FileNotFoundError(f"Could not find model file {model_file}")
        
        if delete_old:
            to_delete = [path.name]
            if metadata_path.exists():
                to_delete.append(metadata_path.name)

            for n in to_delete:
                try:
                    self.api.create_commit(
                        self.name,
                        operations=[CommitOperationDelete(path_in_repo=n,is_folder=False)],
                        commit_message=f"Delete old file: '{n}'",
                        )
                except Exception as e:
                    logging.error(f"Could not delete old file {n} from repo {self.name}")
            

        upload_operations = [CommitOperationAdd(path_in_repo=path.name, path_or_fileobj=path)] 
        if metadata_path.exists():
            upload_operations.append(CommitOperationAdd(path_in_repo=metadata_path.name, path_or_fileobj=metadata_path))
        
        self.api.create_commit(
            self.name,
            upload_operations,
            commit_message=f"Upload new model file: '{path.name}'",)




import numpy as np
from ._base import BaseAdapter
from typing import Union,Tuple,Optional,BinaryIO 
from transformers import AutoTokenizer,AutoModelForCausalLM,AutoConfig,BloomForCausalLM
import os
import torch
import struct
import numpy as np
from ...auto import KnownModels

#based on https://github.com/NouamaneTazi/bloomz.cpp/blob/main/convert-hf-to-ggml.py

conv_map = {
    'word_embeddings'       : 'tok_embeddings',
    "word_embeddings_layernorm": 'norm',
        'input_layernorm'        : 'attention_norm',
        'self_attention.query_key_value': 'attention.query_key_value',
        'self_attention.dense':          'attention.wo',
        'post_attention_layernorm': 'ffn_norm',
        'mlp.dense_h_to_4h'           : 'feed_forward.w1',
        'mlp.dense_4h_to_h'           : 'feed_forward.w2',
        'ln_f'                        : 'output_norm',
        'lm_head' : 'output',
        }

class BloomConverter(BaseAdapter):
    model_type:KnownModels=KnownModels.Bloom

    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoTokenizer,AutoModelForCausalLM]:
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
        
        model = BloomForCausalLM.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
            )
        tokenizer_name = pretrained_tokenizer_name_or_path if pretrained_tokenizer_name_or_path else pretrained_model_name_or_path 
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        return config,tokenizer,model
    
    def _write_hyperparameters(self,out_file:BinaryIO):
        hyperparameters = self.config.to_dict()
        hyperparameters["multiple_of"] = 1
        out_file.write(struct.pack("i", hyperparameters["vocab_size"]))
        out_file.write(struct.pack("i", hyperparameters["hidden_size"]))
        out_file.write(struct.pack("i", hyperparameters["multiple_of"]))
        out_file.write(struct.pack("i", hyperparameters["n_head"]))
        out_file.write(struct.pack("i", hyperparameters["n_layer"]))

    def _rename_weights(self, name: str) -> str:
        #Arrrrrg i hate this, but upstream is inconsistent with naming
        nn = name
        if name != "lm_head.weight":
            nn = nn.split(".")[1:]
        else:
            nn = nn.split(".")

        if nn[0] == "h":
            nn[0] = "layers"
            mapped = conv_map[".".join(nn[2:-1])]
            name = ".".join(nn[:2] + [mapped] + nn[-1:])
        else:
            mapped = conv_map[".".join(nn[:-1])]
            name = ".".join([mapped] + nn[-1:])
        return name

    def _transform_weights(self, name: str, weight: torch.Tensor) ->  torch.Tensor:
        if "query_key_value" in name:
            q, k, v = weight.reshape(self.config.n_head, 3, -1).unbind(1)
            return torch.cat([q, k, v], dim=0).reshape_as(weight)
        return weight
    
    def _filter_weights(self, name: str, weight: torch.Tensor) -> bool:
        return name.endswith("attn.masked_bias") or name.endswith(".attn.bias") 